package check

import (
	"fmt"
	"git.sxidc.com/go-tools/utils/reflectutils"
	"git.sxidc.com/go-tools/utils/strutils"
	"github.com/go-playground/locales/zh"
	ut "github.com/go-playground/universal-translator"
	"github.com/go-playground/validator/v10"
	zhTranslations "github.com/go-playground/validator/v10/translations/zh"
	"github.com/pkg/errors"
	"strings"
	"sync"
	"time"
)

const (
	timeNotZeroTag = "timenotzero"
	whenTag        = "when"
)

var validate = validator.New(validator.WithRequiredStructEnabled())
var translator ut.Translator

var whenMapMutex = sync.RWMutex{}
var whenMap = make(map[string][]string)

func init() {
	validate.SetTagName("check")

	zhLocale := zh.New()
	zhTranslator := ut.New(zhLocale)
	trans, _ := zhTranslator.GetTranslator("zh")

	err := zhTranslations.RegisterDefaultTranslations(validate, trans)
	if err != nil {
		panic(err)
	}

	translator = trans

	registerCustomTags()
}

func registerCustomTags() {
	if err := validate.RegisterValidation(timeNotZeroTag, func(fl validator.FieldLevel) bool {
		fieldValue := fl.Field()

		if !fieldValue.IsValid() {
			return true
		}

		switch value := fieldValue.Interface().(type) {
		case time.Time:
			return !value.IsZero()
		case *time.Time:
			if value != nil {
				return true
			}

			return !value.IsZero()
		default:
			return false
		}
	}); err != nil {
		fmt.Println(err)
		return
	}

	if err := validate.RegisterValidation(whenTag, func(fl validator.FieldLevel) bool {
		whenMapMutex.Lock()
		defer whenMapMutex.Unlock()

		param := strings.Trim(fl.Param(), "/")

		if strutils.IsStringEmpty(param) {
			return true
		}

		structName := reflectutils.PointerValueElem(fl.Top()).Type().Name()
		structFieldName := structName + "." + fl.StructFieldName()
		whenMap[structFieldName] = strings.Split(param, "/")

		return true
	}); err != nil {
		fmt.Println(err)
		return
	}
}

func Struct(obj any, fieldNameMap map[string]string) Result {
	return newResult(validate.Struct(obj), nil, fieldNameMap)
}

type Result struct {
	err            error
	translatedErrs map[string]string
	fieldNameMap   map[string]string
}

func newResult(err error, translatedErrs map[string]string, fieldNameMap map[string]string) Result {
	return Result{
		err:            err,
		translatedErrs: translatedErrs,
		fieldNameMap:   fieldNameMap,
	}
}

func (result Result) IsError() bool {
	return result.err != nil
}

func (result Result) CheckWhen(when string) error {
	return result.CheckFieldWhen(when, nil)
}

func (result Result) CheckFieldWhen(when string, check func(fieldName string) bool) error {
	if result.err == nil {
		return nil
	}

	translatedResult := result.translation()

	if translatedResult.translatedErrs == nil || len(translatedResult.translatedErrs) == 0 {
		return translatedResult.err
	}

	for errStructNamespace, translatedErr := range translatedResult.translatedErrs {
		whenMapMutex.RLock()
		whens := whenMap[errStructNamespace]
		whenMapMutex.RUnlock()

		if strutils.IsStringNotEmpty(when) {
			find := false
			for _, fieldWhen := range whens {
				if fieldWhen == when {
					find = true
					break
				}
			}

			if !find {
				continue
			}
		}

		if check != nil && !check(strings.Split(errStructNamespace, ".")[1]) {
			continue
		}

		if translatedResult.fieldNameMap != nil {
			fieldCNName, ok := translatedResult.fieldNameMap[errStructNamespace]
			if ok {
				translatedErr = strings.ReplaceAll(translatedErr, errStructNamespace, fieldCNName)
			}
		}

		return errors.New(translatedErr)
	}

	return nil
}

func (result Result) translation() Result {
	if result.err == nil {
		return result
	}

	if result.translatedErrs != nil {
		return result
	}

	var validationErrors validator.ValidationErrors
	if !errors.As(result.err, &validationErrors) {
		return result
	}

	translatedErrors := validationErrors.Translate(translator)

	for _, validationError := range validationErrors {
		if validationError.Tag() == timeNotZeroTag {
			translatedErrors[validationError.StructNamespace()] = validationError.Field() + "使用了时间零值"
		}
	}

	return newResult(result.err, translatedErrors, result.fieldNameMap)
}
